<?php

namespace Services;

use Exception\CustomException;
use Util\InstanceProvider;
use Util\Messages;
use Util\RepositoryInstance;
use Util\Utilities;
use Exception\Login as Exception;
use Zend\Mime\Part as ZendMimePart;
use Zend\Mime\Message as ZendMimeMessage;
use Teknisa\Libs\Util\Utilities as LibUtilities;

class Password {
    const BASE_SUBJECT_EMAIL = 'Recuperação de Acesso';
    const USES_ACTIVE_DIRECTORY_POSITIVE = 'Y';
    const USES_PASSWORD_CONTROL_POSITIVE = 'Y';

    const UPPERCASE_TYPE = 'UPPERCASE';
    const NUMERIC_TYPE   = 'NUMERIC';
    const SPECIAL_TYPE   = 'SPECIAL';

    /**
     * @param $email
     * @param $style
     * @throws CustomException
     */
    public function requestNewPassword($email, $style) {
        $userData = InstanceProvider::getLibServiceApi()->getUserData($email, null);
        if(empty($userData) || empty($userData['USER'])) {
            Exception::userNotFound();
        }
        $user = $userData['USER'];
        $email = $user['EMAIL'];
        $passwordParams = $userData['PASSWORD_PARAMETERS'];
        $this->validateExistsAdmUser($passwordParams);
        $password = Utilities::getRandomPassword();
        InstanceProvider::getLibServiceApi()->updatePassword($email, $password, 'R');
        $return = $this->updateNewPasswordVPD($email, 'R');
        $this->sendEmail($user['ID'], $email, $password, $style);
        return $return;
    }

    public function updateNewPasswordVPD($email, $identification) {
        $return = '';
        try {
            $operator = InstanceProvider::getLibUserDataRepository()->getOperatorByNmAccessOrEmail($email);
            $connectionParameter = Utilities::getConnectionParamsParameter();
            if(!empty($operator) && !empty($connectionParameter) && $connectionParameter['useVpd']){
                $operatorSaas = RepositoryInstance::getOperador()->getOperatorByNmAccessOrEmail($email,$operator['ORGANIZATION_ID']);
               if($operatorSaas['PASSWORD_MD5'] !== $operator['PASSWORD_MD5']){
                   RepositoryInstance::getOperador()->updateOperatorPasswordSaas($operator['ID'],
                       $operator['ORGANIZATION_ID'], $operator['PASSWORD_MD5'], $operator['PASSWORD_CRYPTO_PLUS']);
                   RepositoryInstance::getOperador()->insertControlSenha($operator['ID'], $operator['ORGANIZATION_ID'],
                       $operator['PASSWORD_CRYPTO_PLUS'], $operator['PASSWORD_MD5'], $identification);

               }

            }
            $return = 'OK';
        } catch (\Exception $e) {
            $return = $e->getMessage();
        }
        return $return;
    }

    /**
     * @param $actualPassword
     * @param $newPassword
     * @param $newPasswordConfirmation
     * @param $email
     * @throws CustomException
     */
    public function updatePassword($actualPassword, $newPassword, $newPasswordConfirmation, $email) {
        $userData = InstanceProvider::getLibServiceApi()->getUserData($email, null);
        if(empty($userData) || empty($userData['USER'])) {
            Exception::userNotFound();
        }
        $user = $userData['USER'];
        $email = $user['EMAIL'];
        $passwordParams = $userData['PASSWORD_PARAMETERS'];
        $this->validateExistsAdmUser($passwordParams);
        $this->validatePassword($user, $actualPassword, $newPassword, $newPasswordConfirmation, $user['ORGANIZATION_ID']);
        InstanceProvider::getLibServiceApi()->updatePassword($email, $newPassword, 'S');
        InstanceProvider::getSession()->set('EXPIRED_PASSWORD', false);
        $this->updateNewPasswordVPD($email,'S');
    }

    /**
     * @param $operator
     * @param $actualPassword
     * @param $newPassword
     * @param $newPasswordConfirmation
     * @param $organizationId
     * @throws \Exception
     */
    private function validatePassword($operator, $actualPassword, $newPassword, $newPasswordConfirmation, $organizationId) {
        $actPassWeb    = LibUtilities::encrypt($actualPassword);
        $actPassDelphi = LibUtilities::getDelphiPasswordEncrypted($actualPassword);

        if($actPassWeb != $operator['PASSWORD_MD5'] && $actPassDelphi != $operator['PASSWORD_CRYPTO_PLUS']) {
            throw new CustomException('ACTUAL_PASSWORD_NOT_EQUAL');
        }
        if($newPassword != $newPasswordConfirmation) {
            throw new CustomException('NEW_PASSWORD_NOT_EQUAL');
        }

        $this->validatePasswordControlParameters($newPassword, $organizationId, $operator['ID']);
    }

    /**
     * @param $password
     * @param $organizationId
     * @param $userId
     * @throws \Exception
     */
    public function validatePasswordControlParameters($password, $organizationId, $userId) {
        $entityManager = InstanceProvider::getLibConnectionCustom()->updateVpdConnection($organizationId);
        $paramSenha = $this->findPasswordParameters($organizationId, $entityManager);
        if(!empty($paramSenha)) {
            $mask = $paramSenha[0]['CDPICTSENHA'];
            if ($paramSenha[0]['USE_CONTROLSENHA'] == self::USES_PASSWORD_CONTROL_POSITIVE && !empty($mask)) {
                $length    = $this->validatePasswordLength($mask, $password);
                $upperCase = $this->validatePasswordMask($mask, $password, '[M', '/[A-Z]/', self::UPPERCASE_TYPE);
                $numeric   = $this->validatePasswordMask($mask, $password, '[N', '/[0-9]/', self::NUMERIC_TYPE);
                $special   = $this->validatePasswordMask($mask, $password, '[E', '/[!"#$%&()*+,-.:;<=>?@^{|}~]/', self::SPECIAL_TYPE);
                $this->passwordMaskValidation($length, $upperCase, $numeric, $special);
            }
            $changesToReuse = $paramSenha[0]['NRALTSENHREP'];
            if(!empty($changesToReuse)) {
                $latestPassUp = RepositoryInstance::getOperador()->getLatestPasswordUpdates($userId, $changesToReuse, $entityManager);
                foreach($latestPassUp as $latestUp) {
                    if(LibUtilities::encrypt($password) == $latestUp['CDSENHALTWEB']) {
                        Exception::passwordNotReused($changesToReuse);
                    }
                }
            }
        }
    }

    /**
     * @param $length
     * @param $upperCase
     * @param $numeric
     * @param $special
     * @throws CustomException
     */
    private function passwordMaskValidation($length, $upperCase, $numeric, $special) {
        if($length['INVALID_MASK'] || $upperCase['INVALID_MASK'] || $numeric['INVALID_MASK'] || $special['INVALID_MASK']) {
            $params[] = $length['EXISTS_MASK']    ? $length    : null;
            $params[] = $upperCase['EXISTS_MASK'] ? $upperCase : null;
            $params[] = $numeric['EXISTS_MASK']   ? $numeric   : null;
            $params[] = $special['EXISTS_MASK']   ? $special   : null;
            throw new CustomException('INVALID_PASSWORD_MASK', 500, $params);
        }
    }

    /**
     * @param $mask
     * @param $password
     * @return bool
     */
    private function validatePasswordLength($mask, $password) {
        $validation = array (
            'EXISTS_MASK'  => false,
            'INVALID_MASK' => false
        );
        $lengthCaracters = explode('[', $mask)[0];
        if(!empty($lengthCaracters)) {
            $validation['EXISTS_MASK']  = true;
            $validation['MESSAGE']      = 'PASSWORD_VALIDATION_LENGTH';
            $validation['PARAMS']       = array('count' => strlen($lengthCaracters));
            $validation['INVALID_MASK'] = strlen($password) < strlen($lengthCaracters);
        }
        return $validation;
    }

    /**
     * @param $mask
     * @param $password
     * @param $maskStr
     * @param $regexValidation
     * @param $type
     * @return bool
     */
    private function validatePasswordMask($mask, $password, $maskStr, $regexValidation, $type) {
        $validation = array (
            'EXISTS_MASK'  => false,
            'INVALID_MASK' => false
        );
        $maskCount = $this->getMaskCaractersLength($mask, $maskStr);
        if($maskCount) {
            $passwordCount = preg_match_all($regexValidation, $password);
            $validation['EXISTS_MASK']  = true;
            $validation['MESSAGE']      = $this->getPasswordMaskMessage($type);
            $validation['PARAMS']       = array('count' => $maskCount);
            $validation['INVALID_MASK'] = $passwordCount < $maskCount;
        }
        return $validation;
    }

    /**
     * @param $type
     * @return string
     */
    private function getPasswordMaskMessage($type) {
        switch($type) {
            case self::UPPERCASE_TYPE:
                return 'PASSWORD_VALIDATION_UPPERCASE';
                break;
            case self::NUMERIC_TYPE:
                return 'PASSWORD_VALIDATION_NUMERIC';
                break;
            case self::SPECIAL_TYPE:
                return 'PASSWORD_VALIDATION_SPECIAL';
                break;
        }
    }

    /**
     * @param $mask
     * @param $searchStr
     * @return bool|string
     */
    private function getMaskCaractersLength($mask, $searchStr) {
        $length  = false;
        $initPos = strpos($mask, $searchStr);
        if($initPos !== false) {
            $initPos += 2;
            $endPos = strpos($mask, ']', $initPos);
            $length = substr($mask, $initPos, $endPos - $initPos);
        }
        return $length;
    }

    /**
     * @param $style
     * @return string
     */
    private function getResetPasswordTemplatePath($style) {
        $baseDir = __DIR__.'/../Util/EmailTemplates/';
        $customFile = $baseDir . 'reset_password_' . $style . '.html';
        $defaultFile = $baseDir . 'reset_password.html';
        return file_exists($customFile) ? $customFile : $defaultFile;
    }

    /**
     * @param $cdOperator
     * @param $email
     * @param $password
     * @param $style
     * @return mixed|string
     */
    private function getReplacedTemplate($cdOperator, $email, $password, $style) {
        $replacedContents = utf8_decode(file_get_contents($this->getResetPasswordTemplatePath($style)));
        $replacedContents = str_replace('%codigo%', $cdOperator, $replacedContents);
        $replacedContents = str_replace('%email%', $email, $replacedContents);
        $replacedContents = str_replace('%senha%', $password, $replacedContents);
        return $replacedContents;
    }

    /**
     * @param $cdOperator
     * @param $email
     * @param $password
     * @param $style
     */
    private function sendEmail($cdOperator, $email, $password, $style)
    {
        $body     = utf8_encode($this->getReplacedTemplate($cdOperator, $email, $password, $style));
        $smtpMail = InstanceProvider::getSmtpMail();
        $smtpMail->setSubject(self::BASE_SUBJECT_EMAIL);
        $smtpMail->setIsHTML(true);
        $smtpMail->sendMail($body, $email);
    }

    /**
     * @param $passwordParams
     * @throws \Exception
     */
    private function validateExistsAdmUser($passwordParams) {
        if(isset($passwordParams['USES_ACTIVE_DIRECTORY']) &&
                 $passwordParams['USES_ACTIVE_DIRECTORY'] == self::USES_ACTIVE_DIRECTORY_POSITIVE &&
                !$passwordParams['ADM_USER_ACTIVE_DIRECTORY']) {
            throw new CustomException('NOT_ADM_USER_REQUEST_NEW_PASSWORD');
        }
    }

    /**
     * @param $organizationId
     * @param bool $entityManager
     * @return array
     */
    public function findPasswordParameters($organizationId = false, $entityManager = false) {
        $organizationId = $organizationId ? $organizationId : InstanceProvider::getLibEnvironment()->getCurrentOrganizationId();
        return RepositoryInstance::getParamSenha()->getParamSenhaByNrOrg($organizationId, $entityManager);
    }
}